/*
 * Copyright (C) 2012-2019 Michael Hanrath <Michael.Hanrath@uni-koeln.de>
 *
 * This file is part of Diag2PS.
 *
 * Diag2PS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Diag2PS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diag2PS.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "DiagramElements.H"
#include "Parser/Parser.H"
#include <cmath>
#include <fstream>
#include <stdlib.h>
#include <cassert>

namespace Diag2PS { // library namespaces $$$ begin

using namespace std;



DiagramElements::DiagramElements(
	double unitSize,
	double lineWidth, 
	double arrowWidth
	) :
	_state(
		unitSize, lineWidth, "line", arrowWidth, 1.5, "black", "black", 0,
		3.5, // labelDist in pt
		2.0, // pointDist (PSTricks, ncline) in pt
//		-5.83333/2	// labelYShift in pt, half size of a "+"
		-4.30554/2	// labelYShift in pt, half size of an "a"
	),
	_grid(std::make_pair(PosVector(0, 0), PosVector(0, 0)))
{
	_os.setf(std::ios::fixed);
}

DiagramElements::~DiagramElements()
{}


void	DiagramElements::setunitSize(double unitSize)
{	_state.unitSize = unitSize;	}

double	DiagramElements::getunitSize() const
{	return _state.unitSize;	}

void	DiagramElements::setlineWidth(double lineWidth)
{
	_state.lineWidth = lineWidth;
	doSetlineWidth(lineWidth);
}

void	DiagramElements::setbentArrows(bool flag)
{	_state.bentArrows = flag;	}

void	DiagramElements::setarrowFillColor(string color)
{	_state.arrowFillcolor = color;	}

void	DiagramElements::storeBBXCenter(double x)
{	_BBXCenter = x;	}

void	DiagramElements::storeBBYCenter(double y)
{	_BBYCenter = y;	}

void	DiagramElements::storelabelDist(double x)
{	_state.labelDist = x;	}

void	DiagramElements::storepointDist(double x)
{	_state.pointDist = x;	}

void	DiagramElements::storelabelYShift(double y)
{	_state.labelYShift = y;	}


void	DiagramElements::setpath(string color, double width)
{
	_state.pathcolor = color;
	_state.pathWidth = width;
}

void	DiagramElements::setarrowWidth(double arrowWidth)
{	_state.arrowWidth = arrowWidth;	}

void	DiagramElements::setarrowLength2Width(double arrowLength2Width)
{	_state.arrowLength2Width = arrowLength2Width;	}


void	DiagramElements::showGrid(PosVector const & loc1, PosVector const & loc2)
{
	_grid.first = loc1;
	_grid.second = loc2;
}

void	DiagramElements::preamble(const string & s)
{	_preamble += s + "\n";	}	



std::vector<std::pair<std::string, PosVector>>	DiagramElements::loop(
	PosVector const & loc1, 
	LoopFragment loopFragment,
	PosVector const & loc2, 
	double loopWidth,
	ArrowDirection arrowDirection,
	double shift,
	LineMultiplicity lineMultiplicity,
	std::vector<std::pair<std::string, std::vector<Label>>> const & vis
	)
{
std::vector<std::pair<std::string, PosVector>>	vsl;
	if ( loopFragment==FullLoop )
	{
		if ( vis.size()!=2 )
		{
			cerr << "label sequences for full loops must have length 2:" << endl;
			cerr << Diag2PS::currentLine << endl;
			throw FullLoopLabelSequenceLengthError();
		}
		if ( vis[0].second.size() && vis[0].second[0] )
		{
			cerr << "no labels on full loops allowed:" << endl;
			cerr << Diag2PS::currentLine << endl;
			throw NoLabelsOnFullLoops();
		}
		doLoopFragment(loc1, RightLoop, loc2, loopWidth, arrowDirection, shift, lineMultiplicity, std::vector<std::pair<std::string, std::vector<Label>>>(2));
		doLoopFragment(loc1, LeftLoop, loc2, loopWidth, static_cast<ArrowDirection>(-arrowDirection), shift, lineMultiplicity, std::vector<std::pair<std::string, std::vector<Label>>>(2));
	}
	else
	{
	std::vector<PosVector>	vl(doLoopFragment(loc1, loopFragment, loc2, loopWidth, arrowDirection, shift, lineMultiplicity, vis));
	
		assert(vl.size()==vis.size());

		for ( size_t i=0 ; i<vl.size() ; ++i )
//			if ( vis[i].second.size()>0 )	// do not store empty strings as labels
			vsl.push_back(make_pair(vis[i].first, vl[i]));
	}
	
	return vsl;
}




void	DiagramElements::setcolor(string color)
{
	_state.color = color;
}



void	DiagramElements::clear()
{
	_os.str(string());
}


DiagramElements::State const &	DiagramElements::getState() const
{	return _state;	}

DiagramElements &	DiagramElements::setState(State const & state)
{
	_state = state;
	return *this;
}


} // library namespaces $$$ end
