/*
 * Copyright (C) 2012-2019 Michael Hanrath <Michael.Hanrath@uni-koeln.de>
 *
 * This file is part of Diag2PS.
 *
 * Diag2PS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Diag2PS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diag2PS.  If not, see <https://www.gnu.org/licenses/>.
 */


#pragma once

#include <iostream>
#include <sstream>
#include <vector>
#include <cmath>
#include <optional>

#include "Geometry/PosVector.H"
#include "Geometry/DirVector.H"
#include "Label.H"

namespace Diag2PS { // library namespaces $$$ begin

using namespace std;




class DiagramElements {
public:
enum LineType { particleHole, interaction, cluster };
enum LoopFragment { FullLoop, LeftLoop, RightLoop, RightQuarterLoop, LeftQuarterLoop };
enum BubblePos { LeftBubble, RightBubble };
enum ArrowDirection { ArrowBackward3=-3, ArrowBackward2=-2, ArrowBackward=-1, ArrowNone=0, ArrowForward=1, ArrowForward2=2, ArrowForward3=3 };
enum LineMultiplicity { SingleLine, DoubleLine };

	struct FullLoopLabelSequenceLengthError {};
	struct NoLabelsOnFullLoops {};
	
	struct State {
		State(
			double unitSize, double lineWidth, 
			string arrowFillcolor, double arrowWidth, double arrowLength2Width, 
			string color,
			string pathcolor, double pathWidth,
			double labelDist, double pointDist, double labelYShift
			) :
			unitSize(unitSize), lineWidth(lineWidth), 
			arrowFillcolor(arrowFillcolor), arrowWidth(arrowWidth), arrowLength2Width(arrowLength2Width), 
			color(color),
			reverse(false),
			bentArrows(true),
			pathcolor(pathcolor), pathWidth(pathWidth),
			labelDist(labelDist), pointDist(pointDist), labelYShift(labelYShift)
			{}
	
	
		string getEffectivearrowFillColor() const
		{
			if ( arrowFillcolor=="line" )
				return color;
			else
				return arrowFillcolor;
		}
	
	double unitSize;
	double lineWidth;
	string arrowFillcolor;
	double arrowWidth;
	double arrowLength2Width;
	string	color;
	bool	reverse;
	bool	bentArrows;
	string	pathcolor;
	double	pathWidth;
	double	labelDist;	// Diag2PS internal, in pt
	double	pointDist;	// PSTricks labelsep, in pt
	double	labelYShift;	// in pt
	};


	DiagramElements(
		double unitSize = 1.0, // height of loop
		double lineWidth = 1.0,
		double arrowSize = 3	// points
		);
		
	virtual ~DiagramElements();



	
	void	setunitSize(double unitSize);	// height of loop
	double	getunitSize() const;	// height of loop
	void	setlineWidth(double lineWidth);
	void	setarrowWidth(double arrowWidth);
	void	setarrowFillColor(string color);
	void	setbentArrows(bool);
	void	storeBBXCenter(double x);
	void	storeBBYCenter(double y);

	void	storelabelDist(double x);
	void	storepointDist(double x);
	void	storelabelYShift(double y);

	
	void	setarrowLength2Width(double arrowLength2Width);
	void	setcolor(string color);
	void	setpath(string color, double width);

	void	showGrid(PosVector const & loc1, PosVector const & loc2);


	virtual void	verbatim(const string & s) = 0;	
	void	preamble(const string & s);	
	virtual void	rput(PosVector const & loc, const string & s) = 0;	
	virtual void	rput(PosVector const & loc, string const & rputOptions, const string & s) = 0;	
	virtual void	uput(PosVector const & loc, const string & s) = 0;	
	virtual void	uput(PosVector const & loc, string const & uputOptions, const string & s) = 0;	
	virtual void	pointer(PosVector const & loc1, PosVector const & loc2, std::vector<Label> const & labels) = 0;	


	virtual void	doSetlineWidth(double lineWidth) = 0;
	virtual void	cross(PosVector const & loc) = 0;
	virtual void	bullet(PosVector const & loc) = 0;
	virtual void	bbullet(PosVector const & loc) = 0;
	virtual void	circle(PosVector const & loc) = 0;
	virtual void	bcircle(PosVector const & loc) = 0;
	virtual void	disconnectingCircle(PosVector const & loc) = 0;
	virtual void	square(PosVector const & loc) = 0;
	virtual void	bsquare(PosVector const & loc) = 0;

		

	std::vector<std::pair<std::string, PosVector>>	loop(
		PosVector const & loc1, 
		LoopFragment loopFragment,
		PosVector const & loc2, 
		double loopWidth,
		ArrowDirection arrowDirection,
		double shift,
		LineMultiplicity lineMultiplicity,
		std::vector<std::pair<std::string, std::vector<Label>>> const & vis
		);

	virtual void	phLine(
		const PosVector & a, 
		const PosVector & b, 
		ArrowDirection arrowDirection, 
		double arrowShift,
		LineMultiplicity lineMultiplicity,
		std::vector<Label> const & labels
		) = 0;

	virtual void	bubble(
		const PosVector & a, 
		double loopWidth,
		ArrowDirection arrowDirection, 
		LineMultiplicity lineMultiplicity,
		std::vector<Label> const & labels
		) = 0;

	virtual void	snake(
		const PosVector & loc1, 
		LoopFragment loopFragment, 
		const PosVector & loc2, 
		double loopWidth,
		ArrowDirection arrowDirection, 
		double arrowShift,
		LineMultiplicity lineMultiplicity
		) = 0;

	virtual void	interactionLine(const PosVector & a, const PosVector & b) = 0;
	virtual void	interactionDots(const PosVector & a, const PosVector & b) = 0;
	virtual void	photonLine(const PosVector & a, const PosVector & b) = 0;
	virtual void	springLine(const PosVector & a, const PosVector & b) = 0;

	virtual void	clusterLine(const PosVector & a, const PosVector & b) = 0;
	virtual void	clusterLineOpen(const PosVector & a, const PosVector & b) = 0;


	virtual void draw() const = 0;

	DiagramElements & reverse()
	{	_state.reverse = !_state.reverse; return *this;	}

	ostringstream const & sbuf() const
	{	return _os;	}

	void	clear();
	
	
	State const &	getState() const;
	DiagramElements &	setState(State const &);

static constexpr double points_per_cm = 28.45274;

protected:
	virtual std::vector<PosVector>	doLoopFragment(
		PosVector const & loc1, 
		LoopFragment loopFragment, 
		PosVector const & loc2, 
		double loopWidth,
		ArrowDirection arrowDirection,
		double shift,
		LineMultiplicity lineMultiplicity,
		std::vector<std::pair<std::string, std::vector<Label>>> const & yBranches
	) = 0;

	double scale() const
	{	return 1.0/(points_per_cm*_state.unitSize);	}


ostringstream _os;
State	_state;
std::pair<PosVector, PosVector>	_grid;
std::optional<double>	_BBXCenter;
std::optional<double>	_BBYCenter;
std::string	_preamble;
};


} // library namespaces $$$ end
