/*
 * Copyright (C) 2012-2019 Michael Hanrath <Michael.Hanrath@uni-koeln.de>
 *
 * This file is part of Diag2PS.
 *
 * Diag2PS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Diag2PS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diag2PS.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "DiagramInterpreter.H"
#include "DiagramElements.H"
#include "Parser/Parser.H"
#include <fstream>
#include <cassert>
#include <stdlib.h>
#include <unistd.h>

namespace Diag2PS { // library namespaces $$$ begin



DiagramInterpreter::DiagramInterpreter(
	string const & tmpDir,
	string * const pConfigFile, 
	istream & is, 
	DiagramElements * de
	) :
	_de(de), _lastDirection(0, 0)
{
string	tmpin = tmpDir + "/input";
string	tmpout = tmpDir + "/input.preprocessed";
	
	
	{
	ofstream	h(tmpin);
		if ( pConfigFile )
			h << "#include \"" << *pConfigFile << "\"\n";
		h << is.rdbuf();
	}

ostringstream	command;
	command << "gcc -c -I. -E -x c-header -nostdinc " << tmpin << " -o" << tmpout;
//	cerr << command.str() << endl;
	if ( system(command.str().c_str()) )
	{
		cerr << "system command \"" << command.str() << "\" failed" << endl;
		throw PreprocessorExecutionFailed();
	}

	
ifstream	isf(tmpout);
	
	
	
::Diag2PS::Parser driver(*this, de);
//	driver.trace_parsing = true;
//	driver.trace_scanning = true;

	driver.parse(isf);

	if ( driver.errors>0 )
	{
		cerr << driver.errors << " errors" << endl;
		throw ParseErrors();
	}
//	cerr << "no errors" << endl;

//	if ( _de->sbuf().str().size()>0 )
		_de->draw();	// write out anything not saved so far

	*this = driver._diagramInterpreter;


}

void	DiagramInterpreter::move(string const & direction)
{	_state.location += DirVector(direction);	}

void	DiagramInterpreter::move(PosVector const & loc)
{	_state.location = loc;	}

void	DiagramInterpreter::storepos(string const & label)
{
	if ( !_state.storedLocations.insert(std::make_pair(label, _state.location)).second )
		throw DuplicateLabel();
}

void	DiagramInterpreter::storepos(std::vector<std::pair<std::string, PosVector>> const & vsl)
{
	for ( auto const & i : vsl )
		if ( i.first.size()>0 )
			if ( !_state.storedLocations.insert(std::make_pair(i.first, i.second)).second )
				throw DuplicateLabel();
	
}




PosVector	DiagramInterpreter::loadpos(string const & label)
{
map<std::string, PosVector>::const_iterator i(_state.storedLocations.find(label));

	if ( i==_state.storedLocations.end() )
		throw NoSuchLabel();

	return i->second;
}

void	DiagramInterpreter::clearpos(std::vector<std::string> const & labels)
{
	if ( labels.size()>0 )
		for ( auto const & i : labels )
		{
		map<std::string, PosVector>::iterator j(_state.storedLocations.find(i));

			if ( j==_state.storedLocations.end() )
				throw NoSuchLabel();

			_state.storedLocations.erase(j);
		}
	else
		_state.storedLocations.clear();
}

DiagramInterpreter::State const &	DiagramInterpreter::getState() const
{	return _state;	}

DiagramInterpreter &	DiagramInterpreter::setState(State const & state)
{
	_state = state;
	return *this;
}


void	DiagramInterpreter::push()
{
	_stateStack.push(make_pair(_de->getState(), _state));
}

void	DiagramInterpreter::pop()
{
	if ( _stateStack.empty() )
		throw EmptyStack();
		
std::pair<DiagramElements::State, State>	fullState = _stateStack.top();
	_de->setState(fullState.first);
	setState(fullState.second);
	_stateStack.pop();
}


void	DiagramInterpreter::moveLastDirection()
{
	_state.location += _lastDirection;
}

void	DiagramInterpreter::storeDirection(PosVector const & dir)
{
	_lastDirection = DirVector(dir, _state.location);
}


} // library namespaces $$$ end
