/*
 * Copyright (C) 2012-2019 Michael Hanrath <Michael.Hanrath@uni-koeln.de>
 *
 * This file is part of Diag2PS.
 *
 * Diag2PS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Diag2PS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diag2PS.  If not, see <https://www.gnu.org/licenses/>.
 */


#include "Circle.H"

using namespace std;

static double sqr(double x)
{	return x*x;	}


namespace Diag2PS { // library namespaces $$$ begin

Circle::Circle(PosVector const & center, double radius) :
	_center(center), _radius(radius)
{}

Circle::Circle(
	PosVector const & p1,
	PosVector const & p2,
	PosVector const & p3
	) :
	_center(0, 0), _radius(0)
{
	// (x-x0)^2 + (y-y0)^2 = r^2
	// ==> (i): pix^2 - 2*pix*x0 + x0^2   +   piy^2 - 2*piy*y0 + y0^2   =   r^2,    i={1, 2, 3}
	// (i)-(j): 2*(pix-pjx) * x0 + 2*(piy-pjy) * y0 = pix^2 + piy^2 - pjx^2 - pjy^2
	// a00 * x0  +  a01 * y0 = b0			a00 = 2*(p1x-p2x), a01 = 2*(p1y-p2y), b0 = p1x^2 + p1y^2 - p2x^2 - p2y^2
	// a10 * x0  +  a11 * y0 = b1			a10 = 2*(p1x-p3x), a11 = 2*(p1y-p3y), b1 = p1x^2 + p1y^2 - p3x^2 - p3y^2
/*
(%i1) A:matrix([a_00, a_01], [a_10, a_11]);
                                [ a_00  a_01 ]
(%o1)                           [            ]
                                [ a_10  a_11 ]
(%i2) b:matrix([b_0], [b_1]);
                                    [ b_0 ]
(%o2)                               [     ]
                                    [ b_1 ]
(%i3) invert(A).b;
               [       a_11 b_0                a_01 b_1        ]
               [ --------------------- - --------------------- ]
               [ a_00 a_11 - a_01 a_10   a_00 a_11 - a_01 a_10 ]
(%o3)          [                                               ]
               [       a_00 b_1                a_10 b_0        ]
               [ --------------------- - --------------------- ]
               [ a_00 a_11 - a_01 a_10   a_00 a_11 - a_01 a_10 ]
*/
double a00 = 2*(p1.x-p2.x);
double a01 = 2*(p1.y-p2.y);
double a10 = 2*(p1.x-p3.x);
double a11 = 2*(p1.y-p3.y);
double b0 = sqr(p1.x) + sqr(p1.y) - sqr(p2.x) - sqr(p2.y);
double b1 = sqr(p1.x) + sqr(p1.y) - sqr(p3.x) - sqr(p3.y);

double d = a00*a11 - a01*a10;
	if ( std::abs(d)<1e-14 )
		throw PointsOnStraightLine();

	_center = PosVector((a11*b0 - a01*b1)/d, (a00*b1 - a10*b0)/d);
	_radius = sqrt(sqr(p1.x-_center.x) + sqr(p1.y-_center.y));
}



std::ostream & operator << (std::ostream & o, Circle const & c)
{
	o << "[" << c.center() << ", " << c.radius() << "]";
	return o;
}


} // library namespaces $$$ end
