/*
 * Copyright (C) 2012-2019 Michael Hanrath <Michael.Hanrath@uni-koeln.de>
 *
 * This file is part of Diag2PS.
 *
 * Diag2PS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Diag2PS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diag2PS.  If not, see <https://www.gnu.org/licenses/>.
 */


#include "DirVector.H"
#include "PosVector.H"
#include "Angle.H"
#include "Vector_Operations.H"

namespace Diag2PS { // library namespaces $$$ begin

void DirVector::update(double lx, double ly, int & n, double & mult)
{
	if ( n!=0 )
	{
		x += mult*n*lx/10;
		y += mult*n*ly/10;
	}
	if ( n==0 )
	{
		x += mult*lx;
		y += mult*ly;
	}
	n = 0;
	mult = 1.0;
}


DirVector::DirVector(Vector const & v) : Vector(v) {}

DirVector::DirVector(double x, double y, bool swap) :
	Vector(swap ? y : x, swap ? x : y) {}

DirVector::DirVector(PosVector const & a, PosVector const & b) :
	Vector(a - b)
{}



DirVector::DirVector(std::string const & direction) :
	Vector(0, 0)
{
const double _dx = 1/3.0;
const double _dy = 1/3.0;

double lx = 0;
double ly = 0;
int n = 0;
double mult = 1.0;
	for ( std::string::size_type i=0 ; i<direction.size() ; ++i )
	{
		switch ( direction[i] ) {
		case 'r':	lx =   +_dx;	ly =      0;	update(lx, ly, n, mult); break;
		case 'l':	lx =   -_dx;	ly =      0;	update(lx, ly, n, mult); break;
		case 'u':	lx =      0;	ly =   +_dy;	update(lx, ly, n, mult); break;
		case 'd':	lx =      0;	ly =   -_dy;	update(lx, ly, n, mult); break;
		case 'R':	lx = +3*_dx;	ly =      0;	update(lx, ly, n, mult); break;
		case 'L':	lx = -3*_dx;	ly =      0;	update(lx, ly, n, mult); break;
		case 'U':	lx =      0;	ly = +3*_dy;	update(lx, ly, n, mult); break;
		case 'D':	lx =      0;	ly = -3*_dy;	update(lx, ly, n, mult); break;
		case '|':	mult *= 3.0/2.0;	break;
		case '.':	++n;	break;
		}
	}
}


DirVector DirVector::rotate(Angle const & w) const
{
double c = cos(w.value());
double s = sin(w.value());
	return DirVector(c*x - s*y, s*x + c*y);
}


DirVector DirVector::rotate(Angle const & w, DirVector const & r) const
{
double c = cos(w.value());
double s = sin(w.value());
	return DirVector(r.x*(c*x - s*y), r.y*(s*x + c*y));
}

DirVector & DirVector::operator *= (double a)
{
	static_cast<Vector &>(*this) *= a;
	return *this;
}

DirVector & DirVector::operator += (DirVector const & l)
{
	static_cast<Vector &>(*this) += l;
	return *this;
}

DirVector & DirVector::operator -= (DirVector const & l)
{
	static_cast<Vector &>(*this) -= l;
	return *this;
}


DirVector DirVector::perp() const
{
DirVector	n(-y, x);
	n *= 1.0 / n.norm2();
	return n;
}






} // library namespaces $$$ end
