/*
 * Copyright (C) 2012-2019 Michael Hanrath <Michael.Hanrath@uni-koeln.de>
 *
 * This file is part of Diag2PS.
 *
 * Diag2PS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Diag2PS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diag2PS.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "DiagramElements_PSTricks.H"
#include "Geometry/Circle.H"
#include "Geometry/Vector_Operations.H"
#include "Geometry/Angle.H"

#include <cmath>
#include <fstream>
#include <sstream>
#include <array>
#include <stdlib.h>
#include <unistd.h>
#include <cassert>
#include <filesystem>

namespace Diag2PS { // library namespaces $$$ begin

using namespace std;


template <typename T>
struct Rect {
	Rect(T x1, T y1, T x2, T y2) : x1(x1), y1(y1), x2(x2), y2(y2) {}

T x1, y1, x2, y2;
};


template <typename T>
ostream & operator << (ostream & o, Rect<T> const & r)
{
	o << r.x1 << " " << r.y1 << " " << r.x2 << " " << r.y2;
	return o;
}

struct BoundingBox {
	BoundingBox(istream &);

	BoundingBox & xCenter(double x);
	BoundingBox & yCenter(double y);

Rect<int>	bb;
Rect<double>	bbHiRes;
};

BoundingBox::BoundingBox(istream & is) :
	bb(0, 0, 0, 0),
	bbHiRes(0, 0, 0, 0)	
{
string line;
	getline(is, line);
	{
	istringstream is(line);
	string s;
		is >> s >> bb.x1 >> bb.y1 >> bb.x2 >> bb.y2;
	}
	getline(is, line);
	{
	istringstream is(line);
	string s;
		is >> s >> bbHiRes.x1 >> bbHiRes.y1 >> bbHiRes.x2 >> bbHiRes.y2;
	}
}

BoundingBox & BoundingBox::xCenter(double x)
{
double dx1 = bbHiRes.x1 - x;
double dx2 = bbHiRes.x2 - x;

	if ( std::abs(dx1) > std::abs(dx2) )
		bb.x2 = bbHiRes.x2 = x + std::abs(dx1); 
	else
		bb.x1 = bbHiRes.x1 = x - std::abs(dx2); 
		
	return *this;
}

BoundingBox & BoundingBox::yCenter(double y)
{
double dy1 = bbHiRes.y1 - y;
double dy2 = bbHiRes.y2 - y;

	if ( std::abs(dy1) > std::abs(dy2) )
		bb.y2 = bbHiRes.y2 = y + std::abs(dy1); 
	else
		bb.y1 = bbHiRes.y1 = y - std::abs(dy2); 
		
	return *this;
}



ostream & operator << (ostream & o, BoundingBox const & bb)
{
	o << "%%BoundingBox: " << bb.bb << "\n";
	o << "%%HiResBoundingBox: " << bb.bbHiRes << "\n";
	return o;
}



//-----------------------------------------------------------------------------

 
DiagramElements_PSTricks::DiagramElements_PSTricks(
	string const & tmpDir,
	string const & outputFilename,
	OutputTypes outputType,
	double unitSize,
	double lineWidth, 
	double arrowWidth
	) :
	DiagramElements(unitSize, lineWidth, arrowWidth),
	_tmpDir(tmpDir),
	_outputFilename(outputFilename),
	_outputType(outputType)
{}
	


void DiagramElements_PSTricks::draw() const
{
string outputPath = std::filesystem::path(_outputFilename).parent_path();
string outputFilename = std::filesystem::path(_outputFilename).filename();
string texFile = _tmpDir + "/" + outputFilename + ".tex";
string psFile = _tmpDir + "/" + outputFilename + ".ps";
string pdfFile = _tmpDir + "/" + outputFilename + ".pdf";
string svgFile = _tmpDir + "/" + outputFilename + ".svg";
bool runLaTeX = _outputType.ps || _outputType.pdf || _outputType.svg;
int iRuns = ((_BBXCenter || _BBYCenter) && runLaTeX ? 2 : 1);
double xCenter = 0;
double yCenter = 0;
	for ( int iRun=0 ; iRun<iRuns ; ++iRun )
	{
	ostringstream	os;
		os << 
"\\documentclass{article}\n"
"\\usepackage{amssymb}\n"
"\\usepackage{pstricks}\n"
"\\usepackage{pst-node}\n"
"\\usepackage{pst-coil}\n"
"\\usepackage{color}\n"
"\\usepackage{graphicx}\n"
"\\usepackage{pstricks-add}\n"
"\\usepackage{bm}\n"
"\\usepackage{vmargin}\n"
"  \\setpapersize{A3}\n"
"  \\setmargins{0cm}{0cm}% \n"
"             {29cm}{25cm}% \n"
"             {12pt}{25pt}% \n"
"             {0pt}{30pt}% \n"
"\\hoffset=0pt\n"
"\\textwidth=30cm\n"
"\\voffset=0pt\n"
"\\pagestyle{empty}\n"
	<< _preamble << "\n"
"\\begin{document}\n"
"\\psset{unit=" << _state.unitSize << "cm}\n"
//"\\psset{linewidth=" << _state.lineWidth << "pt}\n"
"\\psset{labelsep=2pt}\n"
"\\ \n\n\\vspace{10cm}\n"
"\\ \n\n%% Preamble End\n"
;

		if ( iRun==0 && iRuns==2 )
			os << "\\rput [b](" 
				<< (_BBXCenter ? _BBXCenter.value() : 0) << ", " 
				<< (_BBYCenter ? _BBYCenter.value() : 0) << ")"
				<< "{\\scalebox{0.001}{.}}\n";
		else
		{
			if ( DirVector(_grid.first, _grid.second).norm2()>0.01 )
				os << "\\psgrid[gridwidth=0.25pt,subgriddiv=3,subgriddots=16](0,0)" << _grid.first << _grid.second << "\n";


			{
			istringstream	buf2(_os.str());
			string	line;
			vector<string>	linespath;
			vector<string>	linesBody;
				while ( getline(buf2, line) )
				{
					if ( line.substr(0, 8)=="%%ROTATE" )
					{
						linespath.push_back(line.substr(8));
						linesBody.push_back(line.substr(8));
					}
					else
					{
						if ( line.substr(0, 6)=="%%PATH" )
							linespath.push_back(line.substr(6));
						else
						{
							if ( line.substr(0, 4)=="\\new" )
								linespath.push_back(line);
							else
								linesBody.push_back(line);
						}
					}
				}

				for ( auto const & i : linespath )
					os << i << "\n";
				for ( auto const & i : linesBody )
					os << i << "\n";


			}
		}


		os << "\n\\end{document}\n";



		if ( runLaTeX )
		{
		std::string	sTmpFileTeX(_tmpDir+"/TmpFile.tex");
		std::string	sTmpFileDVI(_tmpDir+"/TmpFile.dvi");
		std::string	sTmpFileAUX(_tmpDir+"/TmpFile.aux");
		std::string	sTmpFileLOG(_tmpDir+"/TmpFile.log");
		std::string	sTmpFilePS(_tmpDir+"/TmpFile.ps");
		std::string	sTmpFileBBOX(_tmpDir+"/TmpFile.bbox");

			{
			ofstream	fileTeX(sTmpFileTeX.c_str());
				fileTeX << os.str();
			}

		ostringstream command;
			command
				<< "cd " << _tmpDir
				<< " && latex " << sTmpFileTeX << " >/dev/null 2>/dev/null"
				<< " && dvips " << sTmpFileDVI << " >/dev/null 2>/dev/null"
				<< " && gs -q -sPAPERSIZE=a3 -sDEVICE=bbox -dNOPAUSE -dSAFER -dBATCH " << sTmpFilePS << " >/dev/null 2>" << sTmpFileBBOX;


			if ( system(command.str().c_str()) )
			{
				cerr << "command \"" << command.str() << "\" failed" << endl;
				throw LaTeXRunFailed();
			}
			ifstream	if2(sTmpFileBBOX.c_str());
			BoundingBox boundingBox(if2);

				if ( iRun==0 && _BBXCenter )
					xCenter = boundingBox.bbHiRes.x1;

				if ( iRun==0 && _BBYCenter )
					yCenter = boundingBox.bbHiRes.y1;



				if ( iRun+1==iRuns )
				{

					if ( _BBXCenter )
						boundingBox.xCenter(xCenter);
					if ( _BBYCenter )
						boundingBox.yCenter(yCenter);

					{
					ifstream	ifs(sTmpFilePS.c_str());
					ofstream	ofs(psFile);

						while ( ifs )
						{
						string	line;
							getline(ifs, line);
							if ( line.substr(0, 21)!="%%DocumentPaperSizes:" && line.substr(0, 19)!="%%HiResBoundingBox:" )
							{
								if ( line.substr(0, 14)=="%%BoundingBox:" )
									ofs << boundingBox;
								else
									ofs << line << "\n";
							}
						}
					}

					if ( _outputType.pdf || _outputType.svg )
					{
					ostringstream command;
						command
							<< "ps2pdf " << psFile << " " << pdfFile
							<< " && pdfcrop --bbox \"`awk '/%%HiResBoundingBox:/ {print $2,$3,$4,$5}' " << psFile << "`\" " << pdfFile << " 2>/dev/null >/dev/null"
							<< " && mv -f " << (_tmpDir+"/"+outputFilename+"-crop.pdf") << " " << pdfFile;
						if ( system(command.str().c_str()) )
						{
							cerr << "command \"" << command.str() << "\" failed" << endl;
							throw PDFcropRunFailed();
						}
					}

					if ( _outputType.svg )
					{
					ostringstream command;
						command << "pdftocairo " << pdfFile << " -svg " << svgFile;
						if ( system(command.str().c_str()) )
						{
							cerr << "command \"" << command.str() << "\" failed" << endl;
							throw PDF2CairoRunFailed();
						}
					}
					
					if ( _outputType.ps )
					{
					std::filesystem::path p(psFile);
						std::filesystem::remove(outputPath / p.filename());
						std::filesystem::copy(p, outputPath / p.filename());
						std::filesystem::remove(p);
					}

					if ( _outputType.pdf )
					{
					std::filesystem::path p(pdfFile);
						std::filesystem::remove(outputPath / p.filename());
						std::filesystem::copy(p, outputPath / p.filename());
						std::filesystem::remove(p);
					}

					if ( _outputType.svg )
					{
					std::filesystem::path p(svgFile);
						std::filesystem::remove(outputPath / p.filename());
						std::filesystem::copy(p, outputPath / p.filename());
						std::filesystem::remove(p);
					}
				}

		}
		if ( !runLaTeX || _outputType.tex )
		{
			if ( _outputFilename.size()>0 )
			{
			std::filesystem::path p(texFile);
			ofstream o(outputPath / p.filename());
				o << os.str();
			}
			else
				cout << os.str();
		}
	}
}


void	DiagramElements_PSTricks::verbatim(const string & s)
{	_os << s << "\n";	}


void	DiagramElements_PSTricks::doSetlineWidth(double lineWidth)
{
	_os << "\\psset{linewidth=" << lineWidth << "pt}\n";
}






//-------------------------------




std::vector<PosVector>	DiagramElements_PSTricks::doLoopFragment(
	PosVector const & loc1, 
	LoopFragment loopFragment, 
	PosVector const & loc2, 
	double loopWidth,
	ArrowDirection arrowDirection,
	double arrowShift,
	LineMultiplicity lineMultiplicity,
	std::vector<std::pair<std::string, std::vector<Label>>> const & yBranches
	)
{
DirVector dir(loc2 - loc1); 
	if ( _state.reverse )
		arrowDirection = static_cast<ArrowDirection>(-arrowDirection);

bool right = loopFragment==RightQuarterLoop || loopFragment==RightLoop;
bool quarter = loopFragment==RightQuarterLoop || loopFragment==LeftQuarterLoop;

//	cerr << loopWidth << endl;
	// PSTricks and Postscript do allow to draw an ellipsis aligned to the principal axes x and y only
	// solution: draw ellipsis in standard orientation and rotate afterwards

DirVector dirDown(0, -(quarter ? 2 : 1)*dir.norm2());
bool widthDirection = (dir.y<0) == right;

double w = (quarter ? loopWidth/3.0 : dir.norm2()*loopWidth/8.0);
PosVector A(loc1);
PosVector B(loc1 + dirDown);
PosVector C(0.5*(A + B) + w*dirDown.perp());
PosVector M(0, 0);	// center
DirVector r(0, 0);	// "radius", for circle: x=y, for ellipsis: x, y components
bool ellipsis = w>dirDown.norm2()/2;
	if ( ellipsis ) 	// ellipsis?
	{
		M =	A + 0.5*dirDown;	// center is in the middle of A and B
		r = w*dirDown.perp() + 0.5*dirDown;
		r.x = abs(r.x);
		r.y = abs(r.y);
	}
	else
	{
	Circle	circle(A, B, C);	// construct circle from three points
		M = circle.center();
		r = DirVector(circle.radius(), circle.radius());
	}

//	cerr << A << B << C << endl;

Angle w1(B-M);
	if ( quarter )
	{
		w1 = 0;
	DirVector d = C - A;
		if ( !widthDirection )
			d.x = -d.x;
		B = A + d;
	}
Angle w2(A-M);
	if ( !widthDirection )
	{
		swap(w1, w2);
		w1 = Angle(M_PI - w1.value());
		w2 = Angle(M_PI - w2.value());
		M.x -= A.x;
		M.x = -M.x;
		M.x += A.x;
	}

double widthShift = scale()*(lineMultiplicity==SingleLine ? 1 : 3)/2;
double lineShift = _state.lineWidth * widthShift;
double pathShift = (_state.pathWidth) * widthShift;
//	cerr << "A=" << A << ", B=" << B << ", C=" << C << ", M=" << M << ", r=" << r << ", w1=" << w1 << ", w2=" << w2 << ", angle = " << dir.angleA()+90 << endl;

	_os << "%%ROTATE\\psrotate" << A << "{" << Angle(dir).value(AngleMeasure::Degree)+90 << "}{%\n";


int signLeftRight = widthDirection ? +1 : -1;
	if ( _state.pathWidth>0 )
	{
		_os 
			<< "%%PATH\\psellipticarc[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]" 
			<< M << DirVector(r.x+pathShift, r.y+pathShift) << "{" << w1.value(AngleMeasure::Degree) << "}{" << w2.value(AngleMeasure::Degree) << "}%\n";
		_os 
			<< "%%PATH\\psline[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]{c-c}" 
			<< A << A << "%\n";
		_os 
			<< "%%PATH\\psline[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]{c-c}" 
			<< B << B << "%\n";
	}


	_os << "\\psellipticarc[linecolor=" << _state.color << ",doubleline=" << (lineMultiplicity==SingleLine ? "false" : "true") << 
		", linewidth=" << _state.lineWidth << "pt" << "]" 
		<< M << DirVector(r.x+lineShift, r.y+lineShift) << "{" << w1.value(AngleMeasure::Degree) << "}{" << w2.value(AngleMeasure::Degree) << "}%\n";
	_os 
		<< "\\psline[linecolor=" << _state.color 
		<< ", linewidth=" << _state.lineWidth << "pt]{c-c}" 
		<< A << A << "%\n";
	_os 
		<< "\\psline[linecolor=" << _state.color 
		<< ", linewidth=" << _state.lineWidth << "pt]{c-c}" 
		<< B << B << "%\n";



std::vector<PosVector>	branches;
double height = -dir.norm2();
	if ( !widthDirection )
		arrowDirection = static_cast<ArrowDirection>(-arrowDirection);

	for ( size_t i=0 ; i<yBranches.size() ; ++i )
	{
	double y = height*i*(1.0/(yBranches.size()-1)) + A.y;
	double x = M.x + signLeftRight*r.x*sqrt((1.0 - sqr((y - M.y)/r.y)));
	PosVector p(x, y);
		p = A + (p-A).rotate(Angle(dir)+Angle(90, AngleMeasure::Degree));
		branches.push_back(p);

		if ( arrowDirection!=ArrowNone  &&  branches.size()>=1  && static_cast<size_t>(i+1)<yBranches.size())
		{
		double y = height*(i+0.5)*(1.0/(yBranches.size()-1)) + A.y;	// position arrow between to interaction lines
		double x = M.x + signLeftRight*r.x*sqrt((1.0 - sqr((y - M.y)/r.y)));
		
		PosVector p(x, y);
		DirVector q(p - M);
		Angle	angle(Angle(q) + Angle(arrowShift, AngleMeasure::Degree));
			q.x *= 1.0/r.x;
			q.y *= 1.0/r.y;
			if ( _state.bentArrows )
				arrowOnEllipticArc(M, r, angle, arrowDirection, lineMultiplicity, ellipsis);
			else
				arrowOnLine(p, angle, arrowDirection, 0, lineMultiplicity);
		}
	}

	_os << "%%ROTATE}%\n";

	for ( size_t i=0 ; i<yBranches.size() ; ++i )
	{
	double y = height*(i+0.5)*(1.0/(yBranches.size()-1)) + A.y;
	double x = M.x + signLeftRight*r.x*sqrt((1.0 - sqr((y - M.y)/r.y)));
	PosVector p(x, y);
		p = M + (p-M).rotate(Angle(arrowShift, AngleMeasure::Degree));
		p = A + (p-A).rotate(Angle(dir) + Angle(90, AngleMeasure::Degree));
		for ( auto const & j : yBranches[i].second )
			if ( j )
				putLabel(j, {loc1, p, loc1 + dir}, dir);
	}

	return branches;
}

//-------------------------------


void	DiagramElements_PSTricks::putLabel(
	Label const & label, 
	array<PosVector, 3> const & bce, 
	DirVector dir,
	bool isPHLine
	)
{
PosVector P(bce[static_cast<int>(label.position)]);
	dir = (1.0/dir.norm2()) * dir;

DirVector	d(0, 0);
DirVector	orientation(label.orientation);
bool deflt = (orientation.x==0  &&  orientation.y==0);
bool beginEnd = label.position==Position::Begin  ||  label.position==Position::End;
double a = _state.labelDist*scale();

	if ( beginEnd )
	{
		if ( deflt )
			d = (label.position==Position::Begin ? -a : +a) * dir;
		else
			d = a*orientation;
	}
	else
	{
		if ( deflt )
			orientation = DirVector(1, 0);
		d = a*orientation;
	}

	P += d;
double ms = 0.5*scale();	// shift by 0.5 pt
	if ( abs(label.orientation.x)==0 && abs(label.orientation.y)==0 )
		P += (ms*label.adjust/d.norm2())*d;
	if ( abs(label.orientation.x)>0 )
		P += DirVector(ms*label.adjust, 0);
	if ( abs(label.orientation.y)>0 )
		P += DirVector(0, ms*label.adjust);


string align;
	if ( label.center )
	{
		if ( d.x>0 )
			align += "l";
		if ( d.x<0 )
			align += "r";
	}
	else
	{
		if ( d.y==0 )
			P += DirVector(0, _state.labelYShift*scale());
		if ( d.y<0 )
			P += DirVector(0, 3*_state.labelYShift*scale());
		align = "B";

		if ( orientation.x>0 )
			align += "l";
		if ( orientation.x<0 )
			align += "r";
	}
	
	_os << "\\rput";
	if ( align.size()>0 )
		_os << "[" << align << "]";
	_os << P << "{\\color{" << _state.color << "}" << label.label << "}\n";
}


//-------------------------------

void	DiagramElements_PSTricks::phLine(
	PosVector const & loc1, 
	PosVector const & loc2, 
	ArrowDirection arrowDirection,
	double arrowShift,
	LineMultiplicity lineMultiplicity,
	std::vector<Label> const & label
	)
{
	if ( _state.pathWidth>0 )
		_os 
			<< "%%PATH\\psline[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]{c-c}" 
			<< loc1 << loc2 << "\n";
	_os 
		<< "\\psline[linecolor=" << _state.color 
		<< ", linewidth=" << _state.lineWidth << "pt, " << "doubleline=" << (lineMultiplicity==SingleLine ? "false" : "true") << "]" << (lineMultiplicity==SingleLine ? "{c-c}" : "") 
		<< loc1 << loc2 << "\n";

	arrowOnLine(loc1, loc2, arrowDirection, arrowShift, lineMultiplicity);
	

	for ( auto const & j : label )
		if ( j )
		{
		DirVector dir(loc2, loc1);
		PosVector P = loc1 + 0.01*(arrowShift+50)*dir;
			putLabel(j, {loc1, P, loc2}, dir);
		}
}

void	DiagramElements_PSTricks::arrowOnLine(
	PosVector const & loc1, 
	PosVector const & loc2, 
	ArrowDirection arrowDirection,
	double arrowShift,
	LineMultiplicity lineMultiplicity
	)
{
	if ( arrowDirection==ArrowNone )
		return;

	if ( _state.reverse )
		arrowDirection = static_cast<ArrowDirection>(-arrowDirection);

double arrowLength2Width = _state.arrowLength2Width;
double arrowWidth = _state.arrowWidth*scale();
double arrowLength = arrowLength2Width*arrowWidth;

	if ( lineMultiplicity==DoubleLine )
		arrowWidth *= 1.5;

	// subtract line thickness
	arrowWidth -= _state.lineWidth*scale();
	arrowLength -= 0.5*_state.lineWidth*scale();

int dirSign = arrowDirection>0 ? +1 : -1;
double adjust = 0.5*dirSign*arrowLength2Width*0.5*_state.lineWidth*scale();

DirVector	dir(loc2-loc1);
PosVector x(loc1 + (0.01*(arrowShift+50) + adjust/dir.norm2())*dir);
	dir *= dirSign/dir.norm2();

	for ( int j=0 ; j<abs(arrowDirection) ; ++j )
	{
	PosVector x0(x + 0.5*(j-(abs(arrowDirection)-1)/2.0)*arrowLength*dir);
	PosVector xu(x0 + 0.5*arrowLength*dir);
	PosVector x1(x0 - 0.5*arrowLength*dir + 0.5*arrowWidth*dir.perp());
	PosVector x2(x0 - 0.5*arrowLength*dir - 0.5*arrowWidth*dir.perp());
			_os
				<< "\\pspolygon[fillstyle=solid,fillcolor=" << _state.getEffectivearrowFillColor() << ",linecolor=" << _state.color 
				<< ",linewidth=" << _state.lineWidth << "pt]"
				<< xu << x1 << x2 << "\n";
	}
}

void	DiagramElements_PSTricks::arrowOnLine(
	PosVector const & loc, 
	Angle angle, 
	ArrowDirection arrowDirection,
	double arrowShift,
	LineMultiplicity lineMultiplicity
	)
{
DirVector d(0, -1);
	d = d.rotate(angle);
	arrowOnLine(loc - d, loc + d, arrowDirection, arrowShift, lineMultiplicity);
}




static double calc_dPhi(double a, double b, double phi, double l)
{
	// approximates difference of two elliptic integrals by derivative
double r = b;
	if ( b>a )
		swap(b, a);

double k = sqrt(1 - b*b/(a*a));
	k *= sin(phi);
		
	return l/(r*sqrt(1 - k*k));
}



void	DiagramElements_PSTricks::arrowOnEllipticArc(
	PosVector const & center, 
	DirVector r, 
	Angle angle,
	bool forward,
	LineMultiplicity lineMultiplicity,
	bool onEllipsis
	)
{
//bool showGeometry = true;
bool showGeometry = false;

double arrowWidth = _state.arrowWidth*scale();
double arrowLength = _state.arrowLength2Width * arrowWidth;

	if ( lineMultiplicity==DoubleLine )
		arrowWidth *= 1.5;

	// subtract line thickness
	arrowWidth -= _state.lineWidth*scale();
	arrowLength -= 0.5*_state.lineWidth*scale();

double dAngle = calc_dPhi(r.x, r.y, angle.value(), arrowLength);
int dirSign = forward ? +1 : -1;
double adjust = -0.5*dirSign*_state.arrowLength2Width*0.5*_state.lineWidth*scale();
	adjust = calc_dPhi(r.x, r.y, angle.value(), adjust);

	angle += adjust;


vector<vector<PosVector>>	vvP(2);
int nPoints = 20;
	for ( int i=0 ; i<2 ; ++i )	// loop over inner/outer arcs
	{
		for ( int j=0 ; j<nPoints ; ++j )	// loop over points on arc
		{
		double fraction = static_cast<double>(j)/(nPoints-1);	// fraction=0: arrow head, fraction=1: arrow tail
		DirVector dx(DirVector(1, 0).rotate(angle.value() + dirSign*dAngle*(fraction - 0.5), r));	// shift along particle/hole line
			double at = atan2(r.x*r.x*dx.y, r.y*r.y*dx.x);	// get normal angle of normal on ellipsis
		PosVector dr(cos(at), sin(at));	
			dr *= 0.5*arrowWidth/dr.norm2();
			vvP[i].push_back(center + dx + (i ? +1 : -1)*fraction*dr);	// calculate points on arc segment of arrow
		}
	}


	// draw actual arrow
	_os << "\\pscustom[fillstyle=solid,fillcolor=" << _state.getEffectivearrowFillColor() 
		<< ",linecolor=" << _state.color 
		<< ",linewidth=" << _state.lineWidth << "pt]{";
	for ( int i=0 ; i<2 ; ++i )	// loop over inner/outer arcs
	{
		_os << "\\psline";
		if ( i==1 )
			_os << vvP[0][nPoints-1];
		
		for ( int j=0 ; j<nPoints ; ++j )	// loop over points on arc
			_os << vvP[i][i==0 ? j : nPoints-(j+1)];
		_os << "%\n";
	}
	_os << "}%\n";



	if ( showGeometry )	// draw foreground objects last
		for ( int i=0 ; i<2 ; ++i )	// loop over inner/outer arcs
			for ( int j=0 ; j<nPoints ; ++j )	// loop over points on arc
				_os << "\\pscircle[fillcolor=black]" << vvP[i][j] << "{0.025}%\n";
}


void	DiagramElements_PSTricks::arrowOnEllipticArc(
	PosVector const & center, 
	DirVector r, 
	Angle angle,
	ArrowDirection arrowDirection,
	LineMultiplicity lineMultiplicity,
	bool onEllipsis
	)
{
double arrowWidth = _state.arrowWidth/points_per_cm;
double arrowLength = _state.arrowLength2Width * arrowWidth;
double l =  0.5*calc_dPhi(r.x, r.y, angle.value(), arrowLength);

	if ( _state.reverse )
		arrowDirection = static_cast<ArrowDirection>(-arrowDirection);
	if ( arrowDirection!=0 )
		for ( int i=0 ; i<abs(arrowDirection) ; ++i )
		{
		double	da = l*(i-abs(arrowDirection/2.0)+0.5);
			da *= arrowDirection>0 ? -1 : +1;
			arrowOnEllipticArc(center, r, angle + da, arrowDirection>0, lineMultiplicity, onEllipsis);
		}
}



void	DiagramElements_PSTricks::bubble(
	const PosVector & loc1, 
	double loopWidth,
	ArrowDirection arrowDirection, 
	LineMultiplicity lineMultiplicity,
	std::vector<Label> const & labels
	)
{
	loopWidth /= 3.0;
PosVector	l(loc1);
DirVector	dir(loopWidth/2.0, 0);
double dx = dir.x;
double r = fabs(dx);
double cm_pt = scale();
double lw = 1*_state.lineWidth/2*cm_pt;
	l += DirVector(dx, 0);
	if ( _state.pathWidth>0 )
		_os 
			<< "%%PATH\\pscircle[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]" 
			<< l << "{" << r+lw << "}";
	_os 
		<< "\\pscircle[linecolor=" << _state.color 
		<< ",linewidth=" << _state.lineWidth << "pt" 
		<< ",doubleline=" << (lineMultiplicity==SingleLine ? "false" : "true") << "]" 
//			<< (lineMultiplicity==SingleLine ? "{c-c}" : "") 
			<< l << "{" << r+lw << "}";

	_os << "\n";

	if ( _state.reverse )
		arrowDirection = static_cast<ArrowDirection>(-arrowDirection);
	if ( _state.bentArrows )
		arrowOnEllipticArc(l, DirVector(r, r), dx>0 ? 0 : M_PI, static_cast<ArrowDirection>(-arrowDirection), lineMultiplicity, false);
	else
		arrowOnLine(loc1 + DirVector(1.85*dx, 0.5), loc1 + DirVector(1.85*dx, -0.5), arrowDirection, 0, lineMultiplicity);

	for ( auto const & i : labels )
		putLabel(i, {loc1, loc1 + 2*dir, loc1}, dir);
}


void	DiagramElements_PSTricks::snake(
	const PosVector & loc1, 
	LoopFragment loopFragment, 
	const PosVector & loc2, 
	double loopWidth,
	ArrowDirection arrowDirection, 
	double arrowShift,
	LineMultiplicity lineMultiplicity
	)
{
DirVector dir(loc2 - loc1);

std::vector<std::pair<std::string, std::vector<Label>>> yBranches(2);
	doLoopFragment(loc1, loopFragment, loc1 + 0.5*dir, loopWidth, arrowDirection, arrowShift, lineMultiplicity, yBranches);
	switch ( loopFragment ) {
		case LeftLoop:		loopFragment = RightLoop;	break;
		case RightLoop:	loopFragment = LeftLoop;	break;
		default:	;
	}
	doLoopFragment(loc1 + 0.5*dir, loopFragment, loc2, loopWidth, arrowDirection, arrowShift, lineMultiplicity, yBranches);
}


void	DiagramElements_PSTricks::cross(PosVector const & loc)
{
	clusterLine(PosVector(loc.x-0.15, loc.y+0.15), PosVector(loc.x+0.15, loc.y-0.15));
	clusterLine(PosVector(loc.x-0.15, loc.y-0.15), PosVector(loc.x+0.15, loc.y+0.15));
}

void	DiagramElements_PSTricks::bullet(PosVector const & loc)
{
	_os << "\\pscircle[fillstyle=solid,fillcolor=" << _state.color << ",linecolor=" << _state.color << "]" << loc << "{0.08}" << "\n";
}

void	DiagramElements_PSTricks::bbullet(PosVector const & loc)
{
	_os << "\\pscircle[fillstyle=solid,fillcolor=" << _state.color << ",linecolor=" << _state.color << "]" << loc << "{0.16}" << "\n";
}

void	DiagramElements_PSTricks::circle(PosVector const & loc)
{
	_os << "\\pscircle[linecolor=" << _state.color << "]" << loc << "{0.07}" << "\n";
}

void	DiagramElements_PSTricks::bcircle(PosVector const & loc)
{
	_os << "\\pscircle[linecolor=" << _state.color << "]" << loc << "{0.15}" << "\n";
}

void	DiagramElements_PSTricks::disconnectingCircle(PosVector const & loc)
{
	_os << "\\pscircle[fillstyle=solid,fillcolor=white,linecolor=" << _state.color << ",linewidth=0.5pt]" << loc << "{0.08}" << "\n";
}

void	DiagramElements_PSTricks::square(PosVector const & loc)
{
const double d = 0.05;
	_os << "\\pspolygon[fillstyle=solid,fillcolor=" << _state.color << ",linecolor=" << _state.color << "]" 
		<< PosVector(loc.x-d, loc.y-d) << PosVector(loc.x-d, loc.y+d) << PosVector(loc.x+d, loc.y+d) << PosVector(loc.x+d, loc.y-d) << PosVector(loc.x-d, loc.y-d) << "\n";
}

void	DiagramElements_PSTricks::bsquare(PosVector const & loc)
{
const double d = 0.10;
	_os << "\\pspolygon[fillstyle=solid,fillcolor=" << _state.color << ",linecolor=" << _state.color << "]" 
		<< PosVector(loc.x-d, loc.y-d) << PosVector(loc.x-d, loc.y+d) << PosVector(loc.x+d, loc.y+d) << PosVector(loc.x+d, loc.y-d) << PosVector(loc.x-d, loc.y-d) << "\n";
}


void	DiagramElements_PSTricks::interactionLine(PosVector const & loc1, PosVector const & loc2)
{
	if ( _state.pathWidth>0 )
		_os 
			<< "%%PATH\\psline[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]{c-c}" 
			<< loc1 << loc2 << "\n";

	_os <<
"\\psline[linewidth=" << 0.75*_state.lineWidth << "pt,linecolor=" << _state.color << ",linestyle=dashed]" << loc1 << loc2 << "\n";
}

void	DiagramElements_PSTricks::interactionDots(PosVector const & loc1, PosVector const & loc2)
{
	if ( _state.pathWidth>0 )
		_os 
			<< "%%PATH\\psline[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]{c-c}" 
			<< loc1 << loc2 << "\n";
	_os <<
"\\psline[linewidth=1.5pt,linecolor=" << _state.color << ",linestyle=dotted]" << loc1 << loc2 << "\n";
}

void	DiagramElements_PSTricks::photonLine(const PosVector & loc1, const PosVector & loc2)
{
double l = DirVector(loc1, loc2).norm2();
DirVector h(loc2, loc1);
Angle angle(h);
	if ( angle.value() < 0 )
		angle += 2*M_PI;
const int periods = 6;
int	li = static_cast<int>(l * periods);
PosVector	hh(loc1);
	hh += DirVector(h.x/(4*li), h.y/(4*li));
	_os << "\\rput{" << angle.value(AngleMeasure::Degree) << "}" << hh << "{\\psCoil[linewidth=1.5pt,linecolor=" 
		<< _state.color << ",coilaspect=0,coilheight=" << (periods*l/li) 
		<< ",coilwidth=" << _state.unitSize/periods << "cm]{-90}{" << li*360-90 << "}}\n";
}

void	DiagramElements_PSTricks::springLine(const PosVector & loc1, const PosVector & loc2)
{
double l = DirVector(loc1, loc2).norm2();
DirVector h(loc2, loc1);
Angle angle(h);
	if ( angle.value() < 0 )
		angle += 2*M_PI;
const int periods = 4;
int	li = static_cast<int>(l * periods);
double	coilaspect = 45;
double	coilheight = (periods*l/li)/cos(coilaspect*M_PI/180.0);
PosVector	hh(loc1);
	hh += DirVector(sin(angle.value())/periods/2, -cos(angle.value())/periods/2);
	_os << "\\rput{" << angle.value(AngleMeasure::Degree) << "}" << hh << "{\\psCoil[linewidth=1.5pt,linecolor=" << _state.color << ",coilaspect=" << coilaspect 
		<< ",coilheight=" << coilheight
		<< ",coilwidth=" << _state.unitSize/periods << "cm]{0}{" << li*360 << "}}\n";
}


void	DiagramElements_PSTricks::clusterLine(PosVector const & loc1, PosVector const & loc2)
{
	if ( _state.pathWidth>0 )
		_os 
			<< "%%PATH\\psline[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]{c-c}" 
			<< loc1 << loc2 << "\n";

	_os <<
"\\psline[linecolor=" << _state.color << ",linewidth=" << 2*_state.lineWidth << "pt]{c-c}" << loc1 << loc2 << "\n";
}

void	DiagramElements_PSTricks::clusterLineOpen(PosVector const & loc1, PosVector const & loc2)
{
	if ( _state.pathWidth>0 )
		_os 
			<< "%%PATH\\psline[linecolor=" << _state.pathcolor 
			<< ", linewidth=" << _state.pathWidth << "pt]{c-c}" 
			<< loc1 << loc2 << "\n";

const double d = 0.03;
	_os <<
"\\psline[linecolor=" << _state.color << ",linewidth=" << 0.5*_state.lineWidth << "pt]" << PosVector(loc1.x, loc1.y-d) << DirVector(loc2.x, loc2.y-d) << "\n";
	_os <<
"\\psline[linecolor=" << _state.color << ",linewidth=" << 0.5*_state.lineWidth << "pt]" << PosVector(loc1.x, loc1.y+d) << DirVector(loc2.x, loc2.y+d) << "\n";
}


void	DiagramElements_PSTricks::rput(PosVector const & loc, const string & s)
{
	_os << "\\rput" << loc << "{\\color{" << _state.color << "}" << s << "}%\n";
}	

void	DiagramElements_PSTricks::rput(PosVector const & loc, string const & rputOptions, const string & s)
{
	_os << "\\rput" << rputOptions << loc << "{\\color{" << _state.color << "}" << s << "}%\n";
}	

void	DiagramElements_PSTricks::uput(PosVector const & loc, const string & s)
{
	_os << "\\uput" << loc << "{\\color{" << _state.color << "}" << s << "}%\n";
}	

void	DiagramElements_PSTricks::uput(PosVector const & loc, string const & uputOptions, const string & s)
{
	_os << "\\uput" << uputOptions << loc << "{\\color{" << _state.color << "}" << s << "}%\n";
}	

void	DiagramElements_PSTricks::pointer(PosVector const & loc1, PosVector const & loc2, std::vector<Label> const & labels)
{
	for ( auto const & i : labels )
	{
	DirVector dir(loc2 - loc1);
		putLabel(i, {loc1, 0.5*(loc1+loc2), loc2}, dir, false);
		dir *= (_state.pointDist*scale()/dir.norm2());
		_os << "\\pnode" << loc2 << "{P}";
		_os << "\\pnode" << loc1+dir << "{Q}";
		
		_os << "\\ncline[linewidth=1.5pt,nodesep=0pt,linecolor=" << _state.color << "]{->}{P}{Q}%\n";
	}
}
	

} // library namespaces $$$ end
